import asyncio
import logging
import queue
import threading
import uuid
from typing import Any
from typing import Protocol

import aiohttp
import pygame

from pongy import settings
from pongy.models import MoveDirection
from pongy.models import WsCommand
from pongy.models import WsCommandMovePayload
from pongy.models import WsErrorEvent
from pongy.models import WsEvent
from pongy.models import WsGameStateEvent
from pongy.ui.widgets.ball import BallWidget
from pongy.ui.widgets.racket import RacketWidgetFactory
from pongy.ui.widgets.score import ScoreWidgetFactory

logger = logging.getLogger(__name__)


class Exit:
    pass


class IApplication(Protocol):
    host: str
    port: int
    input_queue: queue.Queue[Exit | WsEvent]
    output_queue: queue.Queue[WsCommand]


class Connection(threading.Thread):
    def __init__(self, app: IApplication):
        super().__init__(daemon=True)
        self._app = app

    def run(self) -> None:
        self._connect()

    def connect(self) -> None:
        self.start()

    async def _process_output_queue(self, ws: aiohttp.ClientWebSocketResponse) -> None:
        while True:
            await asyncio.sleep(1 / (settings.FPS * 10))
            while not self._app.output_queue.empty():
                await ws.send_json(self._app.output_queue.get_nowait().dict())

    async def _handle(self) -> None:
        player_id: str = str(uuid.uuid4())
        url = f"ws://{self._app.host}:{self._app.port}/ws"
        process_queue_task: asyncio.Task[Any] | None = None
        try:
            async with aiohttp.ClientSession() as session:
                async with session.ws_connect(
                    url,
                    heartbeat=settings.WS_HEARTBEAT_TIMEOUT,
                    headers={"Cookie": f"player_id={player_id}"},
                ) as ws:
                    process_queue_task = asyncio.create_task(
                        self._process_output_queue(ws)
                    )
                    async for msg in ws:
                        if msg.type == aiohttp.WSMsgType.TEXT:
                            ws_event = WsEvent.parse_raw(msg.data)
                            self._app.input_queue.put(ws_event)
        except aiohttp.ClientConnectionError:
            logger.error("Connection error")
        except Exception as err:  # pylint: disable=broad-except
            logger.exception(err)
        else:
            logger.error("Connection lost")
        finally:
            self._app.input_queue.put(Exit())
            if process_queue_task:
                process_queue_task.cancel()

    def _connect(self) -> None:
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self._handle())


class Application:
    def __init__(self, host: str, port: int):
        self.host: str = host
        self.port: int = port
        self.input_queue: queue.Queue[Exit | WsEvent] = queue.Queue()
        self.output_queue: queue.Queue[WsCommand] = queue.Queue()

    def run(self) -> None:
        self._connect()
        pygame.init()
        surface = pygame.display.set_mode((settings.BOARD_SIZE, settings.BOARD_SIZE))
        surface.fill(settings.BOARD_COLOR)
        while True:
            ws_event = self.input_queue.get(block=True)
            if isinstance(ws_event, Exit):
                logger.info("Exit")
                break
            if isinstance(ws_event.data, WsErrorEvent):
                logger.error(ws_event.data.payload.message)
                break
            self._process_key_pressed()
            self._redraw(surface, ws_event.data)
        pygame.quit()

    def _connect(self) -> None:
        connection = Connection(app=self)
        connection.connect()

    def _redraw(
        self, surface: pygame.surface.Surface, ws_event: WsGameStateEvent
    ) -> None:
        surface.fill(settings.BOARD_COLOR)
        for player in ws_event.payload.players:
            racket_widget = RacketWidgetFactory(player.racket.side).create(
                player.racket.position
            )
            score_widget = ScoreWidgetFactory(player.racket.side).create(player.score)
            racket_widget.draw(surface)
            score_widget.draw(surface)

        ball_position = ws_event.payload.ball.position
        BallWidget(ball_position).draw(surface)
        pygame.display.flip()

    def _process_key_pressed(self) -> None:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                logger.debug("Got exit signal")
                self.input_queue.put(Exit())
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_q:
                    logger.debug("Got exit signal")
                    self.input_queue.put(Exit())
        keys = pygame.key.get_pressed()
        if keys[pygame.K_LEFT] or keys[pygame.K_UP]:
            ws_event = WsCommand(
                payload=WsCommandMovePayload(direction=MoveDirection.LEFT)
            )
            self.output_queue.put(ws_event)
        if keys[pygame.K_RIGHT] or keys[pygame.K_DOWN]:
            ws_event = WsCommand(
                payload=WsCommandMovePayload(direction=MoveDirection.RIGHT)
            )
            self.output_queue.put(ws_event)



import asyncio
import logging
from types import TracebackType
from typing import Any

from pongy import settings
from pongy.models import BoardSide
from pongy.models import WsEvent
from pongy.models import WsGameStateEvent
from pongy.models import WsGameStatePayload
from pongy.server.ball import Ball
from pongy.server.ball import IBall
from pongy.server.player import IPlayer
from pongy.server.racket import BottomRacket
from pongy.server.racket import IRacket
from pongy.server.racket import LeftRacket
from pongy.server.racket import RightRacket
from pongy.server.racket import TopRacket

logger = logging.getLogger(__name__)


class Game:
    def __init__(self) -> None:
        self.available_rackets: list[IRacket] = [
            RightRacket(),
            LeftRacket(),
            TopRacket(),
            BottomRacket(),
        ]
        self.players: list[IPlayer] = []
        self.ball: IBall = Ball()
        self._run_task: asyncio.Task[Any] = asyncio.create_task(self.run())

    def add_player(self, player: IPlayer) -> None:
        try:
            player.racket = self.available_rackets.pop()
        except IndexError:
            logger.error("No available racket error")
        self.players.append(player)
        logger.debug("Added new player")

    def remove_player(self, player: IPlayer) -> None:
        player.racket.reset()
        self.available_rackets.append(player.racket)
        self.players[:] = [p for p in self.players if p.uuid != player.uuid]
        logger.debug("Removed player")
        if self.is_empty:
            self._run_task.cancel()

    def bounce_notify(self, side: BoardSide) -> None:
        for player in self.players:
            player.bounce_notify(side)

    def bounce(self) -> None:
        new_x, new_y = self.ball.position
        if new_x < 0:
            new_x = 0
            self.ball.angle = 180 - self.ball.angle
            self.bounce_notify(BoardSide.LEFT)
        elif new_x > settings.BOARD_SIZE - settings.BALL_SIZE:
            new_x = settings.BOARD_SIZE - settings.BALL_SIZE
            self.ball.angle = 180 - self.ball.angle
            self.bounce_notify(BoardSide.RIGHT)
        if new_y < 0:
            new_y = 0
            self.ball.angle = -self.ball.angle
            self.bounce_notify(BoardSide.TOP)
        elif new_y > settings.BOARD_SIZE - settings.BALL_SIZE:
            new_y = settings.BOARD_SIZE - settings.BALL_SIZE
            self.ball.angle = -self.ball.angle
            self.bounce_notify(BoardSide.BOTTOM)
        self.ball.position = int(new_x), int(new_y)

    def to_payload(self) -> WsGameStatePayload:
        return WsGameStatePayload(
            ball=self.ball.to_payload(),
            players=[player.to_payload() for player in self.players],
        )

    async def run(self) -> None:
        while True:
            asyncio.create_task(self.broadcast())
            await asyncio.sleep(1 / settings.FPS)
            self.ball.move()
            for player in self.players:
                player.racket.hit(self.ball)
            self.bounce()

    async def broadcast(self) -> None:
        payload = WsEvent(data=WsGameStateEvent(payload=self.to_payload()))
        await asyncio.gather(
            *(subscriber.ws.send_json(payload.dict()) for subscriber in self.players),
            return_exceptions=True
        )

    @property
    def is_full(self) -> bool:
        return len(self.players) == 4

    @property
    def is_empty(self) -> bool:
        return not self.players


class GamePool:
    _awaiting: Game | None = None

    def __init__(self, player: IPlayer) -> None:
        self._player: IPlayer = player
        self._game: Game | None = None

    async def __aenter__(self) -> Game:
        if not GamePool._awaiting:
            self._game = GamePool._awaiting = Game()
            logger.debug("Created new game")
        else:
            self._game = GamePool._awaiting
        self._game.add_player(self._player)
        if self._game.is_full:
            GamePool._awaiting = None
        return self._game

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._game:
            self._game.remove_player(self._player)
            if GamePool._awaiting is self._game and self._game.is_empty:
                GamePool._awaiting = None



import pygame

from pongy import settings
from pongy.models import BoardSide


class BaseScoreWidget:
    position = (0, 0)

    def __init__(self, score: int):
        self._score = str(score)

    def draw(self, surface: pygame.surface.Surface) -> None:
        font = pygame.font.SysFont("arial", settings.SCORE_TEXT_SIZE)
        text = font.render(self._score, True, settings.SCORE_COLOR)
        surface.blit(text, self.position)


class BottomScoreWidget(BaseScoreWidget):
    position = (
        settings.BOARD_SIZE // 2,
        settings.BOARD_SIZE // 2 + settings.SCORE_TEXT_SHIFT,
    )


class TopScoreWidget(BaseScoreWidget):
    position = (
        settings.BOARD_SIZE // 2,
        settings.BOARD_SIZE // 2 - settings.SCORE_TEXT_SHIFT,
    )


class LeftScoreWidget(BaseScoreWidget):
    position = (
        settings.BOARD_SIZE // 2 - settings.SCORE_TEXT_SHIFT,
        settings.BOARD_SIZE // 2,
    )


class RightScoreWidget(BaseScoreWidget):
    position = (
        settings.BOARD_SIZE // 2 + settings.SCORE_TEXT_SHIFT,
        settings.BOARD_SIZE // 2,
    )


class ScoreWidgetFactory:
    def __init__(self, side: BoardSide):
        self.side = side

    def create(self, score: int) -> BaseScoreWidget:
        mapping = {
            BoardSide.RIGHT: RightScoreWidget,
            BoardSide.LEFT: LeftScoreWidget,
            BoardSide.TOP: TopScoreWidget,
            BoardSide.BOTTOM: BottomScoreWidget,
        }
        return mapping[self.side](score)



import math
from dataclasses import dataclass
from random import randint
from typing import Protocol

from pongy import settings
from pongy.models import WsBall


class IBall(Protocol):
    speed: int
    angle: int
    position: tuple[int, int]

    def move(self) -> None:
        pass

    def change_speed(self) -> None:
        pass

    def to_payload(self) -> WsBall:
        pass


@dataclass
class Ball:
    speed: int = settings.DEFAULT_BALL_SPEED
    angle: int = randint(20, 160)
    position: tuple[int, int] = ((settings.BOARD_SIZE - settings.BALL_SIZE) // 2,) * 2

    def move(self) -> None:
        radians = math.radians(self.angle)
        new_x = self.position[0] + self.speed * math.cos(radians)
        new_y = self.position[1] + self.speed * math.sin(radians)
        self.position = int(new_x), int(new_y)

    def change_speed(self) -> None:
        self.speed = randint(settings.MIN_BALL_SPPED, settings.MAX_BALL_SPEED)

    def to_payload(self) -> WsBall:
        return WsBall(position=self.position)



from dataclasses import dataclass
from random import randint
from typing import Protocol

from pongy import settings
from pongy.models import BoardSide
from pongy.models import MoveDirection
from pongy.models import WsRacket
from pongy.server.ball import IBall


class IRacket(Protocol):
    position: int
    side: BoardSide

    def hit(self, ball: IBall) -> None:
        pass

    def move(self, direction: MoveDirection) -> None:
        pass

    def reset(self) -> None:
        pass

    def to_payload(self) -> WsRacket:
        pass


@dataclass
class BaseRacket:
    position: int = (settings.BOARD_SIZE - settings.RACKET_LENGTH) // 2
    side: BoardSide = BoardSide.BOTTOM

    def hit(self, ball: IBall) -> None:
        pass

    def reset(self) -> None:
        self.position = (settings.BOARD_SIZE - settings.RACKET_LENGTH) // 2

    def move(self, direction: MoveDirection) -> None:
        if direction == MoveDirection.LEFT:
            if self.position - settings.RACKET_SPEED >= 0:
                self.position -= settings.RACKET_SPEED
            else:
                self.position = 0
        if direction == MoveDirection.RIGHT:
            if (
                self.position + settings.RACKET_SPEED
                <= settings.BOARD_SIZE - settings.RACKET_LENGTH
            ):
                self.position += settings.RACKET_SPEED
            else:
                self.position = settings.BOARD_SIZE - settings.RACKET_LENGTH

    def to_payload(self) -> WsRacket:
        return WsRacket(
            position=self.position,
            side=self.side,
        )


@dataclass
class BottomRacket(BaseRacket):
    side: BoardSide = BoardSide.BOTTOM

    def hit(self, ball: IBall) -> None:
        new_x, new_y = ball.position
        if (
            new_y > settings.BOARD_SIZE - settings.BALL_SIZE - settings.RACKET_HEIGHT
            and self.position - settings.BALL_SIZE
            < new_x
            < self.position + settings.RACKET_LENGTH + settings.BALL_SIZE
        ):
            new_y = settings.BOARD_SIZE - settings.BALL_SIZE - settings.RACKET_HEIGHT
            ball.angle = randint(200, 340)
            ball.change_speed()
        ball.position = int(new_x), int(new_y)


@dataclass
class TopRacket(BaseRacket):
    side: BoardSide = BoardSide.TOP

    def hit(self, ball: IBall) -> None:
        new_x, new_y = ball.position
        if (
            new_y < settings.RACKET_HEIGHT
            and self.position - settings.BALL_SIZE
            < new_x
            < self.position + settings.RACKET_LENGTH + settings.BALL_SIZE
        ):
            new_y = settings.RACKET_HEIGHT
            ball.angle = randint(20, 160)
            ball.change_speed()
        ball.position = int(new_x), int(new_y)


@dataclass
class LeftRacket(BaseRacket):
    side: BoardSide = BoardSide.LEFT

    def hit(self, ball: IBall) -> None:
        new_x, new_y = ball.position
        if (
            new_x < settings.RACKET_HEIGHT
            and self.position - settings.BALL_SIZE
            < new_y
            < self.position + settings.RACKET_LENGTH + settings.BALL_SIZE
        ):
            new_x = settings.RACKET_HEIGHT
            ball.angle = randint(-70, 70)
            ball.change_speed()
        ball.position = int(new_x), int(new_y)


@dataclass
class RightRacket(BaseRacket):
    side: BoardSide = BoardSide.RIGHT

    def hit(self, ball: IBall) -> None:
        new_x, new_y = ball.position
        if (
            new_x > settings.BOARD_SIZE - settings.BALL_SIZE - settings.RACKET_HEIGHT
            and self.position - settings.BALL_SIZE
            < new_y
            < self.position + settings.RACKET_LENGTH + settings.BALL_SIZE
        ):
            new_x = settings.BOARD_SIZE - settings.BALL_SIZE - settings.RACKET_HEIGHT
            ball.angle = randint(110, 250)
            ball.change_speed()
        ball.position = int(new_x), int(new_y)



from dataclasses import dataclass
from typing import Protocol

from aiohttp import web

from pongy.models import BoardSide
from pongy.models import WsPlayer
from pongy.server.racket import BaseRacket
from pongy.server.racket import IRacket


class IPlayer(Protocol):
    uuid: str
    ws: web.WebSocketResponse
    score: int
    racket: IRacket

    def bounce_notify(self, side: BoardSide) -> None:
        pass

    def to_payload(self) -> WsPlayer:
        pass


@dataclass
class Player:
    uuid: str
    ws: web.WebSocketResponse
    score: int = 0
    racket: IRacket = BaseRacket()

    def bounce_notify(self, side: BoardSide) -> None:
        if self.racket.side == side:
            self.score += 1

    def to_payload(self) -> WsPlayer:
        return WsPlayer(
            uuid=self.uuid, score=self.score, racket=self.racket.to_payload()
        )
